# -*- encoding: utf-8 -*-

if (Gem::Specification.find_all_by_name 'oily_png') != []
  require 'oily_png'
else
  require 'chunky_png'
end

include ChunkyPNG::Color
#==How to use:
#
#===1.Running from a command line
# ruby cli_run.rb -h
#
# or run a test run with
# rake sample_run
#===2.Including this module to a class
#
#  require 'png_comparison'
#
#  class YourClass
#    include PNG_Comparison
#  end
#
#  c = YourClass.new
#  c.compare(img1: image1, img2: image2, diff_file: diff_file, threshold: threshold)
#
# this will make available the method "PNG_Comparison::compare()" to your class. The parameters the same as above
# img1,2,diff_file can be either filenames or instances of ChunkyPNG::Image class
#
#===3.Running compare() method directly from your program
#  require 'png_comparison'
#  ...
#  PNG_Comparison::compare(img1: image1, img2: image2, diff_file: diff_file, threshold: threshold)
#
#==Tips and tricks:
#===For CRuby users: to  speed up the program, please run:
# * "gem install 'oily_png'" in cmd after.
#====this will make comparison run much faster for big images as it uses C extensions. But oily_png gem is only for C platform
#
#===For JRuby users:
# Please adjust JVM memory when comparing big images to avoid OutOfMemory error :
#  jruby -J-Xmx2200m png_comparison.rb [params]
# or use  the above flag  when running your own scripts that will use the png_comparison gem.
# Please bear in mind that the flag should be placed after 'jruby' but before your script name.
# The above memory size should be enough for images like 1700x7200, which is quite a regular web page size today.
#
# The main image difference algorithm is a kind of CIE76 (see https://en.wikipedia.org/wiki/Color_difference#CIE76).

module PNG_Comparison
  module Comparator
    def self.compare_images(img1:, img2:, diff_file:, threshold:)
      read_start = Time.now
      puts 'Reading images...'
      if !(img1.is_a?(ChunkyPNG::Image) && img2.is_a?(ChunkyPNG::Image) && diff_file.is_a?(ChunkyPNG::Image))
        image1 = ChunkyPNG::Image.from_file(img1)
        image2 = ChunkyPNG::Image.from_file(img2)
        output_file = ChunkyPNG::Image.new(image1.width, image1.height, WHITE)
      else
        image1 = img1
        image2 = img2
        output_file = diff_file
        diff_file = 'diff_file'
      end

      read_end = Time.now
      puts "Finished reading images in #{read_end - read_start} seconds"

      image_length = image1.pixels.length
      diff = []
      perceptual_diff = 0

      if !image1.eql?(image2) && (image1.dimension == image2.dimension)
        diff, output_file = img_diff(image1: image1, image2: image2, output_file: output_file)
        perceptual_diff = (diff.inject { |sum, value| sum + value } / image_length) * 100
      else if image1.dimension != image2.dimension
             puts 'Images have different dimensions!'
             return false
        end
      end

      absolute_change = show_statistics(image_length, diff, perceptual_diff, threshold)
      # you can use here "if absolute_change > perceptual_diff"
      if absolute_change > threshold
        puts 'Images are not identical.'
        output_file.save(diff_file)
        return false
      else
        puts 'Images are identical or differ within threshold.'
        output_file.save(diff_file) unless image1.eql?(image2)
        return true
      end
    end

    private

    # Main color difference algorithm, kind of CIE76 metric
    def self.img_diff(image1:, image2:, output_file:)
      diff = []
      image1.height.times do |y|
        image1.row(y).each_with_index do |pixel, x|
          next if pixel == image2[x, y]
          score = Math.sqrt((r(image2[x, y]) - r(pixel))**2 + (g(image2[x, y]) - g(pixel))**2 +
                                       (b(image2[x, y]) - b(pixel))**2) / Math.sqrt(MAX**2 * 3)
          output_file[x, y] = grayscale(MAX - (score * MAX).round)
          diff << score
        end
      end
      [diff, output_file]
    end

    def self.show_statistics(image_length, diff, perceptual_diff, threshold)
      total_pixels = image_length
      non_matching_pixels = diff.length
      absolute_change = (non_matching_pixels.to_f / total_pixels.to_f) * 100
      puts "Pixels processed: #{total_pixels}"
      puts "Differing Pixels: #{non_matching_pixels}"
      puts "Perceptual image difference: #{perceptual_diff} (%)"
      puts "Absolute image difference: #{absolute_change} (%), threshold is #{threshold}(%)"
      absolute_change
    end
  end

  # This method will appear as an instance method if you include PNG_Comparison module to your class.
  # will return true if images are identical, otherwize false
  def compare(img1:, img2:, diff_file:, threshold:)
    begin
      start = Time.now
      comparison = PNG_Comparison::Comparator.compare_images(img1: img1, img2: img2, diff_file: diff_file, threshold: threshold.to_f)
      finish = Time.now
      puts "Comparison took #{finish - start} seconds."
    rescue Exception => e
      puts "This error has occured: #{e.message}"
      puts "The backtrace is #{e.backtrace}"
      return nil
    end
    comparison
  end

  extend self
end
