require_relative '../lib/png_comparison'

describe PNG_Comparison do
  before(:context) do
    @command = if RUBY_PLATFORM == 'java'
                 'jruby '
               else
                 'ruby '
               end
    @image_path = './spec/spec_images/'
    Dir.mkdir(@image_path) unless Dir.exist?(@image_path)
    @image1_file = @image_path + 'valid_base.png'
    @image2_file = @image_path + 'valid_base2.png'
    @wrong_dimension_image_filename = @image_path + 'wrong_dimension.png'
    @diff_file = @image_path + 'diff_file.png'
    @c = Comparison.new
    @valid_image = fill_in(ChunkyPNG::Image.new(200, 200, WHITE), 0)
    @valid_image2 = fill_in(ChunkyPNG::Image.new(200, 200, WHITE), 5)
    @another_dimension_image = fill_in(ChunkyPNG::Image.new(120, 100, WHITE), 50)
    @valid_image.save(@image1_file)
    @valid_image2.save(@image2_file)
    @another_dimension_image.save(@wrong_dimension_image_filename)
  end

  def fill_in(image, offset)
    off = 0
    image.height.times do |y|
      image.row(y).each_with_index do |_pixel, x|
        off = offset if (x > image.width.to_f / 2) & (y > image.height.to_f / 2)
        image[x, y] = rgb(x + off, (x + y) - 10 + off, y)
      end
    end
    image
  end

  class Comparison
    include PNG_Comparison
  end

  describe '#run' do
    context 'from the command line -' do
      it "doesn't save the output when images are the same", same_images_file_command_line: true do
        cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@image1_file} --diff-file #{@diff_file} --threshold 0.5", 'r+')
        output = cons.read
        puts output
        expect(File.exist?(@diff_file)).to be false
      end

      it 'saves the difference to a specified file', different_images_file_command_line: true do
        cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@image2_file} --diff-file #{@diff_file} --threshold 5", 'r+')
        output = cons.read
        puts output
        expect(File.exist?(@diff_file)).to be true
      end

      it 'prints a valid specified threshold to the command line', valid_threshold_command_line: true do
        cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@image2_file} --diff-file #{@diff_file} --threshold 5", 'r+')
        output = cons.read
        puts output
        expect(output).to include('threshold is 5.0(%)')
      end

      it "shows an error when image dimension doesn't match", dim_error_command_line: true do
        cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@wrong_dimension_image_filename} --diff-file #{@diff_file} --threshold 5}", 'r+')
        output = cons.read
        puts output
        expect(output).to include('Images have different dimensions!')
      end

      it "prints 'Images are identical or differ within threshold.'", same_images_true_command_line: true do
        cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@image1_file} --diff-file #{@diff_file} --threshold 0.5", 'r+')
        output = cons.read
        puts output
        expect(output).to include('Images are identical or differ within threshold.')
      end

      context "in case there are less or more than 4 arguments provided, \
                it prints error 'Error: Wrong number of Parameters!!! The arguments should be: \
              [path][img1].png [path][img2].png [path][diff].png [threshold] (=0.05 by default),\
                                  [path] can be absolute or relative' ", wrong_arguments: true do
        it '  for 1 argument', different_images_file_command_line_1arg: true do
          cons = IO.popen(@command + "./cli_run.rb  --file1 #{@image1_file}", 'r+')
          output = cons.read
          puts output
          expect(output).to include('Error: not all the mandatory parameters set. please run with -h/--help option')
        end

        it ' for 2 arguments', different_images_file_command_line_2arg: true do
          cons = IO.popen(@command + "./cli_run.rb  --file1 #{@image1_file}  --file2 #{@image1_file}", 'r+')
          output = cons.read
          puts output
          expect(output).to include('Error: not all the mandatory parameters set. please run with -h/--help option')
        end

        it ' for 3 arguments', different_images_file_command_line_3arg: true do
          cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@image2_file} --diff-file #{@diff_file}", 'r+')
          output = cons.read
          puts output
          expect(output).to include('Error: not all the mandatory parameters set. please run with -h/--help option')
        end

        it ' for 6 arguments', different_images_file_command_line_6arg: true do
          cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file}  --file2 #{@image2_file} --diff-file #{@diff_file} --file1 #{@image1_file}\
                                                                                              --file2 #{@image2_file} -diff-file #{@diff_file}", 'r+')
          output = cons.read
          puts output
          expect(output).to include('Error: not all the mandatory parameters set. please run with -h/--help option')
        end
      end

      it 'shows right statistics', statistics_command_line: true do
        cons = IO.popen(@command + "./cli_run.rb --file1 #{@image1_file} --file2 #{@image2_file} --diff-file #{@diff_file} --threshold 5", 'r+')
        output = cons.read
        puts output
        expect(File.exist?(@diff_file)).to be true
        expect(output).to include('Reading images...')
        expect(output).to include('Finished reading images')
        expect(output).to include('Pixels processed:')
        expect(output).to include('Differing Pixels:')
        expect(output).to include('Perceptual image difference:')
        expect(output).to include('Perceptual image difference:')
        expect(output).to include('Comparison took')
      end

      it "shows help when run with -h key" do
        cons = IO.popen(@command + "./cli_run.rb -h", 'r+')
        output = cons.read
        puts output
        expect(output).to include("The arguments should be:")
        expect(output).to include("--file1 [path/]img1.png  --file2 [path/]img2.png --diff-file [path/]diff.png")
        expect(output).to include("--threshold [threshold in percents], [path] can be absolute or relative")
      end
    end

    context 'when using module method' do
      it 'returns true when images are different (below threshold)', different_images_below_threshold: true do
        expect(PNG_Comparison.compare(img1:@image1_file, img2:@image2_file, diff_file:@diff_file, threshold:60)).to be true
      end

      it "accepts ChunkyPNG::Image data as input parameters" do
        diff_image = ChunkyPNG::Image.new(200, 200, WHITE)
        #require 'pry'
        #binding.pry
        expect(PNG_Comparison.compare(img1:@valid_image, img2:@valid_image2, diff_file:diff_image, threshold:60)).to be true
      end
    end

    context 'included to a class ', same_images: true do
      it 'returns true when images are different (below threshold)', different_images_below_threshold: true do
        expect(@c.compare(img1:@image1_file, img2:@image2_file, diff_file:@diff_file, threshold:60)).to be true
      end

      it 'returns false when images are different (above threshold)', different_images_above_threshold: true do
        expect(@c.compare(img1:@image1_file, img2:@image2_file, diff_file:@diff_file, threshold:5)).to be false
      end
    end
  end

  after(:each) do
    File.delete(@diff_file) if File.exist?(@diff_file)
  end

  after(:context) do
    Dir.glob(@image_path + '*.png').each { |file| File.delete(file) }
    Dir.delete(@image_path)
  end
end
